/*
	BmSmtp.cpp
		- Implements the main SMTP-client-class: BmSmtp

		$Id: BmSmtp.cpp,v 1.34 2003/11/10 16:24:12 zooey Exp $
*/
/*************************************************************************/
/*                                                                       */
/*  Beam - BEware Another Mailer                                         */
/*                                                                       */
/*  http://www.hirschkaefer.de/beam                                      */
/*                                                                       */
/*  Copyright (C) 2002 Oliver Tappe <beam@hirschkaefer.de>               */
/*                                                                       */
/*  This program is free software; you can redistribute it and/or        */
/*  modify it under the terms of the GNU General Public License          */
/*  as published by the Free Software Foundation; either version 2       */
/*  of the License, or (at your option) any later version.               */
/*                                                                       */
/*  This program is distributed in the hope that it will be useful,      */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    */
/*  General Public License for more details.                             */
/*                                                                       */
/*  You should have received a copy of the GNU General Public            */
/*  License along with this program; if not, write to the                */
/*  Free Software Foundation, Inc., 59 Temple Place - Suite 330,         */
/*  Boston, MA  02111-1307, USA.                                         */
/*                                                                       */
/*************************************************************************/


#include <ctype.h>
#include <memory.h>
#include <stdio.h>

#ifdef BEAM_FOR_BONE
# include <netinet/in.h>
#endif
#include <NetAddress.h>
#include <NetEndpoint.h>

#include "regexx.hh"
using namespace regexx;

#include "BmBasics.h"
#include "BmEncoding.h"
	using namespace BmEncoding;
#include "BmJobStatusWin.h"
#include "BmIdentity.h"
#include "BmLogHandler.h"
#include "BmMail.h"
#include "BmMailHeader.h"
#include "BmMemIO.h"
#include "BmNetUtil.h"
#include "BmPopAccount.h"
#include "BmPopper.h"
#include "BmSmtpAccount.h"
#include "BmSmtp.h"
#include "BmPrefs.h"
#include "BmUtil.h"

/********************************************************************************\
	BmSmtpStatusFilter
\********************************************************************************/

/*------------------------------------------------------------------------------*\
	()
		-	
\*------------------------------------------------------------------------------*/
BmSmtpStatusFilter::BmSmtpStatusFilter( BmMemIBuf* input, uint32 blockSize)
	:	inherited( input, blockSize)
	,	mAtStartOfLine( true)
	,	mDigitCount( 0)
	,	mIncludeUpToNewline( false)
{
}

/*------------------------------------------------------------------------------*\
	()
		-	
\*------------------------------------------------------------------------------*/
void BmSmtpStatusFilter::Reset( BmMemIBuf* input) {
	inherited::Reset( input);
	mAtStartOfLine = true;
	mDigitCount = 0;
	mIncludeUpToNewline = false;
}

/*------------------------------------------------------------------------------*\
	()
		-	
\*------------------------------------------------------------------------------*/
void BmSmtpStatusFilter::Filter( const char* srcBuf, uint32& srcLen, 
											char* destBuf, uint32& destLen) {
	const char* src = srcBuf;
	const char* srcEnd = srcBuf+srcLen;

	if (mHaveStatus) {
		uint32 size = min( destLen, srcLen);
		memcpy( destBuf, srcBuf, size);
		srcLen = destLen = size;
	} else {
		for( ;src<srcEnd; ++src) {
			if (mIncludeUpToNewline) {
				if (*src=='\n') {
					mHaveStatus = true;
					break;
				} else
					continue;
			} else if (mAtStartOfLine || mDigitCount) {
				if (mDigitCount==3 && *src!='-') {
					mIncludeUpToNewline = true;
				} else if (mDigitCount<3 && isdigit( *src))
					mDigitCount++;
				else
					mDigitCount = 0;
			}
			mAtStartOfLine = (*src=='\n');
		}
		uint32 statusSize = src-srcBuf;
		mStatusText.Append( srcBuf, statusSize);
		if (mHaveStatus) {
			src++;								// skip newline
			if (!mNeedData)
				mEndReached = true;
		}
		srcLen = src-srcBuf;
		destLen = 0;
	}
}

/*------------------------------------------------------------------------------*\
	()
		-	
\*------------------------------------------------------------------------------*/
bool BmSmtpStatusFilter::CheckForPositiveAnswer() {
	if (mStatusText.Length() && mStatusText[0] > '3') {
		BmString err("Server answers: \n");
		err += mStatusText;
		err.RemoveAll( "\r");
		throw BM_network_error( err);
	}
	return true;
}



/********************************************************************************\
	BmSmtp
\********************************************************************************/

// standard logfile-name for this class:
#undef BM_LOGNAME
#define BM_LOGNAME Name()

const char* const BmSmtp::MSG_SMTP = 		"bm:smtp";
const char* const BmSmtp::MSG_DELTA = 		"bm:delta";
const char* const BmSmtp::MSG_TRAILING = 	"bm:trailing";
const char* const BmSmtp::MSG_LEADING = 	"bm:leading";

// message component definitions for additional info:
const char* const BmSmtp::MSG_PWD = 	"bm:pwd";

// job-specifier for checking server capabilities:
const int32 BmSmtp::BM_CHECK_CAPABILITIES_JOB = 1;

/*------------------------------------------------------------------------------*\
	SmtpStates[]
		-	array of SMTP-states, each with title and corresponding handler-method
\*------------------------------------------------------------------------------*/
BmSmtp::SmtpState BmSmtp::SmtpStates[BmSmtp::SMTP_FINAL] = {
	SmtpState( "auth via pop...", &BmSmtp::StateAuthViaPopServer),
	SmtpState( "connect...", &BmSmtp::StateConnect),
	SmtpState( "helo...", &BmSmtp::StateHelo),
	SmtpState( "auth...", &BmSmtp::StateAuth),
	SmtpState( "send...", &BmSmtp::StateSendMails),
	SmtpState( "quit...", &BmSmtp::StateDisconnect),
	SmtpState( "done", NULL)
};

/*------------------------------------------------------------------------------*\
	BmSmtp( info)
		-	contructor
\*------------------------------------------------------------------------------*/
BmSmtp::BmSmtp( const BmString& name, BmSmtpAccount* account)
	:	inherited( BmString("SMTP_")<<name, BM_LogSmtp, 
					  new BmSmtpStatusFilter( NULL))
	,	mSmtpAccount( account)
	,	mMsgTotalSize( 0)
	,	mCurrMailNr( 0)
	,	mState( 0)
	,	mServerMayHaveSizeLimit( false)
	,	mServerSupportsDSN( false)
	,	mPopAccAcquisitorFunc( NULL)
{
}

/*------------------------------------------------------------------------------*\
	~BmSmtp()
		-	destructor
		-	frees all associated memory (hopefully)
\*------------------------------------------------------------------------------*/
BmSmtp::~BmSmtp() { 
	TheLogHandler->FinishLog( BM_LOGNAME);
}

/*------------------------------------------------------------------------------*\
	ShouldContinue()
		-	determines whether or not the Smtp-job should continue to run
		-	in addition to the inherited behaviour, the Smtp-job will continue
			when it executes special jobs (not BM_DEFAULT_JOB), since in that
			case there are no controllers present.
\*------------------------------------------------------------------------------*/
bool BmSmtp::ShouldContinue() {
	return inherited::ShouldContinue() 
			 || CurrentJobSpecifier() == BM_CHECK_CAPABILITIES_JOB;
}

/*------------------------------------------------------------------------------*\
	StartJob()
		-	the mainloop, steps through all SMTP-stages and calls the 
			corresponding handlers
\*------------------------------------------------------------------------------*/
bool BmSmtp::StartJob() {
	int32 startState = SMTP_CONNECT;
	if (mSmtpAccount->NeedsAuthViaPopServer()
	&& CurrentJobSpecifier() != BM_CHECK_CAPABILITIES_JOB)
		// authentication through pop-server takes place before anything else:
		startState = SMTP_AUTH_VIA_POP;

	const float delta = 100.0 / (SMTP_DONE-startState);
	const bool failed=true;
	try {
		for( 	mState = startState; 
				ShouldContinue() && mState<SMTP_DONE; ++mState) {
			if (mState == SMTP_AUTH && mSmtpAccount->NeedsAuthViaPopServer())
				continue;
			TStateMethod stateFunc = SmtpStates[mState].func;
			UpdateSMTPStatus( (mState==startState ? 0.0 : delta), NULL);
			(this->*stateFunc)();
			if (CurrentJobSpecifier() == BM_CHECK_CAPABILITIES_JOB 
			&& mState==SMTP_HELO) {
				// CHECK-CAPABILITIES-job is done after HELO:
				UpdateSMTPStatus( delta*(SMTP_DONE-mState), NULL);
				return true;
			}
		}
		if (!ShouldContinue())
			UpdateSMTPStatus( 0.0, NULL, false, true);
		else
			UpdateSMTPStatus( delta, NULL);
		mSmtpAccount->mMailVect.clear();
	}
	catch( BM_runtime_error &err) {
		// a problem occurred, we tell the user:
		BmString errstr = err.what();
		int e;
		if (mConnection && (e = mConnection->Error())!=B_OK)
			errstr << "\nerror: " << e << ", " << mConnection->ErrorStr();
		UpdateSMTPStatus( 0.0, NULL, failed);
		BmString text = Name() << ":\n\n" << errstr;
		HandleError( text);
		mSmtpAccount->mMailVect.clear();
		return false;
	}
	return true;
}

/*------------------------------------------------------------------------------*\
	UpdateSMTPStatus( delta, detailText, failed)
		-	informs the interested party about a change in the current SMTP-state
		-	failed==true means that we only want to indicate the failure of the
			current stage (the BmString "FAILED!" will be shown)
\*------------------------------------------------------------------------------*/
void BmSmtp::UpdateSMTPStatus( const float delta, const char* detailText, 
										  bool failed, bool stopped) {
	auto_ptr<BMessage> msg( new BMessage( BM_JOB_UPDATE_STATE));
	msg->AddString( MSG_SMTP, Name().String());
	msg->AddString( BmJobModel::MSG_DOMAIN, "statbar");
	msg->AddFloat( MSG_DELTA, delta);
	if (failed)
		msg->AddString( MSG_TRAILING, 
							 (BmString(SmtpStates[mState].text) 
								<< " FAILED!").String());
	else if (stopped)
		msg->AddString( MSG_TRAILING, 
							 (BmString(SmtpStates[mState].text) 
							 	<< " Stopped!").String());
	else
		msg->AddString( MSG_TRAILING, SmtpStates[mState].text);
	if (detailText)
		msg->AddString( MSG_LEADING, detailText);
	TellControllers( msg.get());
}

/*------------------------------------------------------------------------------*\
	UpdateMailStatus( delta, detailText)
		- informs the interested party about the message currently dealt with
\*------------------------------------------------------------------------------*/
void BmSmtp::UpdateMailStatus( const float delta, const char* detailText, 
										 int32 currMsg) {
	BmString text = BmString() << currMsg << " of " 
							<< mSmtpAccount->mMailVect.size();
	auto_ptr<BMessage> msg( new BMessage( BM_JOB_UPDATE_STATE));
	msg->AddString( MSG_SMTP, Name().String());
	msg->AddString( BmJobModel::MSG_DOMAIN, "mailbar");
	msg->AddFloat( MSG_DELTA, delta);
	msg->AddString( MSG_LEADING, text.String());
	if (detailText)
		msg->AddString( MSG_TRAILING, detailText);
	TellControllers( msg.get());
}

/*------------------------------------------------------------------------------*\
	UpdateProgress( numBytes)
		-
\*------------------------------------------------------------------------------*/
void BmSmtp::UpdateProgress( uint32 numBytes) {
	float delta = (100.0 * numBytes) / (mMsgTotalSize ? mMsgTotalSize : 1);
	BmString detailText = BmString("size: ") << BytesToString( mCurrMailSize);
	UpdateMailStatus( delta, detailText.String(), mCurrMailNr);
}

/*------------------------------------------------------------------------------*\
	StateConnect()
		-	Initiates network-connection to SMTP-server
\*------------------------------------------------------------------------------*/
void BmSmtp::StateConnect() {
	BNetAddress addr;
	if (!mSmtpAccount->GetSMTPAddress( &addr)) {
		BmString s = BmString("Could not determine address of SMTP-Server ") 
							<< mSmtpAccount->SMTPServer();
		throw BM_network_error( s);
	}
	if (!Connect( &addr)) {
		BmString s = BmString("Could not connect to SMTP-Server ") 
							<< mSmtpAccount->SMTPServer() 
						 	<< "\n\bError:\n\t"<< mErrorString;
		throw BM_network_error( s);
	}
	CheckForPositiveAnswer();
}

/*------------------------------------------------------------------------------*\
	StateHelo()
		-	Sends greeting to server and checks result
		-	EHLO is tried first, aiming to find out more about server 
			capabilities; if that fails, HELO is used
		-	if EHLO succeeds, info about the server-capabilities are extracted
			from the answer
\*------------------------------------------------------------------------------*/
void BmSmtp::StateHelo() {
	BmString domain = mSmtpAccount->DomainToAnnounce();
	if (!domain.Length())
		domain = OwnFQDN();
	BmString cmd = BmString("EHLO ") << domain;
	SendCommand( cmd);
	try {
		CheckForPositiveAnswer( true);
							// we expect a multiline string as answer
		Regexx rx;
		if (rx.exec( StatusText(), "^\\d\\d\\d.SIZE\\b", Regexx::newline)) {
			mServerMayHaveSizeLimit = true;
		}
		if (rx.exec( StatusText(), "^\\d\\d\\d.DSN\\b", Regexx::newline)) {
			mServerSupportsDSN = true;
		}
		if (rx.exec( StatusText(), "^\\d\\d\\d.AUTH\\W+(.*?)$", Regexx::newline)) {
			mSupportedAuthTypes = rx.match[0].atom[0];
		}
	} catch(...) {
		cmd = BmString("HELO ") << domain;
		SendCommand( cmd);
		CheckForPositiveAnswer();
	}
}

/*------------------------------------------------------------------------------*\
	StateAuthViaPopServer()
		-	authenticates through pop-server (SMTP_AFTER_POP)
\*------------------------------------------------------------------------------*/
void BmSmtp::StateAuthViaPopServer() {
	BmMail* mail = mSmtpAccount->mMailVect[0].Get();
	if (mail) {
		BmString sender = mail->Header()->DetermineSender();
		BmString accName = mSmtpAccount->AccForSmtpAfterPop();
		BmRef<BmPopAccount> sendingAcc = dynamic_cast<BmPopAccount*>( 
			ThePopAccountList->FindItemByKey( accName).Get()
		);
		if (!sendingAcc) {
			// no default pop-account set, we try to find out by looking at the
			// sender address:
			sendingAcc = TheIdentityList->FindPopAccountForAddrSpec( sender);
		}
		if (!sendingAcc) {
			// still no pop-account found, we ask user, if possible:
			if (mPopAccAcquisitorFunc) {
				if (!mPopAccAcquisitorFunc( Name(), accName)) {
					Disconnect();
					StopJob();
					return;
				}
				sendingAcc = dynamic_cast<BmPopAccount*>( 
					ThePopAccountList->FindItemByKey( accName).Get()
				);
			}
		}
		if (!sendingAcc)
			BM_THROW_RUNTIME( BmString("Sorry, could not determine pop-account "
											"for address '") << sender 
										<< "'\n\n Smtp-after-Pop authentication failed, "
											"giving up.");
		BmRef<BmPopper> popper( new BmPopper( 
			sendingAcc->Name(), 
			sendingAcc.Get()
		));
		popper->SetPwdAcquisitorFunc( BmPopperView::AskUserForPwd);
		popper->StartJobInThisThread( BmPopper::BM_AUTH_ONLY_JOB);
	}
}

/*------------------------------------------------------------------------------*\
	StateAuth()
		-	Sends user/passwd combination and checks result
		-	currently supports PLAIN- and LOGIN-authentication
\*------------------------------------------------------------------------------*/
void BmSmtp::StateAuth() {
	BmString pwd;
	bool pwdOK = false;
	bool first = true;
	BmString authMethod = mSmtpAccount->AuthMethod();
	authMethod.ToUpper();
	if (!authMethod.Length())
		return;			// no authentication needed...
	while(!pwdOK) {
		if (first && mSmtpAccount->PwdStoredOnDisk()) {
			pwd = mSmtpAccount->Password();
		} else if (mPwdAcquisitorFunc) {
			if (!ShouldContinue() || !mPwdAcquisitorFunc( Name(), pwd)) {
				Disconnect();
				StopJob();
				return;
			}
		} else
			BM_THROW_RUNTIME( "Unable to acquire password !?!");

		first = false;
		if (authMethod == BmSmtpAccount::AUTH_PLAIN) {
			// PLAIN-method: send single base64-encoded string that is composed
			// like this:
			// authenticateID + \0 + username + \0 + password
			// (where authenticateID is currently always empty)
			BmString cmd = BmString("AUTH PLAIN");
			SendCommand( cmd);
			if (CheckForPositiveAnswer()) {
				BmString base64;
				cmd = BmString("_") << mSmtpAccount->Username() << "_" 
							<< mSmtpAccount->Password();
				int32 len=cmd.Length();
				char* buf = cmd.LockBuffer( len);
				buf[0] = '\0';
				buf[mSmtpAccount->Username().Length()+1] = '\0';
				cmd.UnlockBuffer( len);
				Encode( "base64", cmd, base64);
				SendCommand( "", base64);
			}
		} else if (authMethod == BmSmtpAccount::AUTH_LOGIN) {
			// LOGIN-method: send base64-encoded username, then send 
			// base64-encoded password:
			BmString cmd = BmString("AUTH LOGIN");
			SendCommand( cmd);
			if (CheckForPositiveAnswer()) {
				BmString base64;
				Encode( "base64", mSmtpAccount->Username(), base64);
				SendCommand( base64);
				CheckForPositiveAnswer();
				Encode( "base64", mSmtpAccount->Password(), base64);
				SendCommand( "", base64);
			}
		}
		try {
			pwdOK = CheckForPositiveAnswer();
		} catch( BM_network_error &err) {
			// most probably a wrong password...
			BmString errstr = err.what();
			int e;
			if (mConnection && (e = mConnection->Error())!=B_OK)
				errstr << "\nerror: " << e << ", " << mConnection->ErrorStr();
			BmString text = Name() << ":\n\n" << errstr;
			HandleError( text);
		}
	}
}

/*------------------------------------------------------------------------------*\
	StateSendMails()
		-	sends the queued mails to the server (invokes MAIL, RCPT and 
			DATA-commands)
		-	depending on the handling of Bcc-recipients, each mail is only sent to
			the server once (SpecialHeaderForEachBcc=false) or each mail is being
			sent once for the standard recipients (To, Cc) plus a personalized 
			version for each Bcc-recipient (SpecialHeaderForEachBcc=true)
\*------------------------------------------------------------------------------*/
void BmSmtp::StateSendMails() {
	Regexx rx;
	mMsgTotalSize = 0;
	for( uint32 i=0; i<mSmtpAccount->mMailVect.size(); ++i) {
		BmMail* mail = mSmtpAccount->mMailVect[i].Get();
		if (mail)
			mMsgTotalSize += mail->RawText().Length();
	}

	mCurrMailNr = 1;
	for( uint32 i=0; i<mSmtpAccount->mMailVect.size(); ++i, ++mCurrMailNr) {
		BmMail* mail = mSmtpAccount->mMailVect[i].Get();
		if (!mail) {
			BM_LOGERR( BmString("SendMails(): mail no. ") << i 
								<< " is NULL, skipping it.");
			continue;
		}
		mCurrMailSize = mail->RawText().Length();

		BmString headerText = mail->HeaderText();
		if (!mail->Header()->IsFieldEmpty(BM_FIELD_RESENT_BCC)) {
			// remove RESENT-BCC-header from mailtext...
			headerText = rx.replace( 
				headerText, 
				"^Resent-Bcc:\\s*.+?\\r\\n(\\s+.*?\\r\\n)*", 
				"", Regexx::newline
			);
		}
		if (!mail->Header()->IsFieldEmpty(BM_FIELD_BCC)) {
			// remove BCC-header from mailtext...
			headerText = rx.replace( 
				headerText, 
				"^Bcc:\\s*.+?\\r\\n(\\s+.*?\\r\\n)*", 
				"", Regexx::newline
			);
		}

		BmRcptVect rcptVect;
		if (ThePrefs->GetBool("SpecialHeaderForEachBcc", true)) {
			if (HasStdRcpts( mail, rcptVect)) {
				Mail( mail);
				Rcpt( rcptVect);
				Data( mail, headerText);
			}
			BccRcpt( mail, true, headerText);
		} else {
			Mail( mail);
			if (HasStdRcpts( mail, rcptVect))
				Rcpt( rcptVect);
			BccRcpt( mail, false, headerText);
			Data( mail, headerText);
		}
		if (ShouldContinue())
			mail->MarkAs( BM_MAIL_STATUS_SENT);
	}
	mCurrMailSize = 0;
	mCurrMailNr = 0;
	mSmtpAccount->mMailVect.clear();
}

/*------------------------------------------------------------------------------*\
	StateDisconnect()
		-	tells the server that we are finished
\*------------------------------------------------------------------------------*/
void BmSmtp::StateDisconnect() {
	Quit( true);
}

/*------------------------------------------------------------------------------*\
	Mail( mail)
		-	announces the given mail to the server
\*------------------------------------------------------------------------------*/
void BmSmtp::Mail( BmMail* mail) {
	BmString sender = mail->Header()->DetermineSender();
	Regexx rx;
	if (!rx.exec( sender, "@\\w+")) {
		// no domain part within sender-address, we add our current domain:
		if (sender.FindFirst("@") == B_ERROR)
			sender << "@";
		BmString fqdn = mSmtpAccount->DomainToAnnounce();
		sender << OwnDomain( fqdn);
	}
	BmString cmd = BmString("MAIL from:<") << sender <<">";
	if (mServerMayHaveSizeLimit) {
		int32 mailSize = mail->RawText().Length();
		cmd << " SIZE=" << mailSize;
	}
	SendCommand( cmd);
	CheckForPositiveAnswer();
}

/*------------------------------------------------------------------------------*\
	bool HasStdRcpts( mail)
		-	determines the list of standard-recipients (from TO & CC)
		-	returns true if at least one recipient has been set, false otherwise
			(the latter case might occur if a mail only contains BCCs)
\*------------------------------------------------------------------------------*/
bool BmSmtp::HasStdRcpts( BmMail* mail, BmRcptVect& rcptVect) {
	BmAddrList::const_iterator iter;
	const BmAddressList& toList 
		= mail->IsRedirect() 
			? mail->Header()->GetAddressList( BM_FIELD_RESENT_TO)
			: mail->Header()->GetAddressList( BM_FIELD_TO);
	for( iter=toList.begin(); iter != toList.end(); ++iter) {
		if (!iter->HasAddrSpec())
			// empty group-addresses have no real address-specification 
			// (like 'Undisclosed-Recipients: ;'), we filter those:
			continue;
		rcptVect.push_back( iter->AddrSpec());
	}
	const BmAddressList& ccList 
		= mail->IsRedirect() 
			? mail->Header()->GetAddressList( BM_FIELD_RESENT_CC)
			: mail->Header()->GetAddressList( BM_FIELD_CC);
	for( iter=ccList.begin(); iter != ccList.end(); ++iter) {
		if (!iter->HasAddrSpec())
			// empty group-addresses have no real address-specification 
			// (like 'Undisclosed-Recipients: ;'), we filter those:
			continue;
		rcptVect.push_back( iter->AddrSpec());
	}
	return rcptVect.size()>0;
}

/*------------------------------------------------------------------------------*\
	bool Rcpt( rcptVect)
		-	announces all given recipients to the server
\*------------------------------------------------------------------------------*/
void BmSmtp::Rcpt( const BmRcptVect& rcptVect) {
	for( uint32 i=0; i<rcptVect.size(); ++i) {
		BmString cmd = BmString("RCPT to:<") << rcptVect[i] <<">";
		SendCommand( cmd);
		CheckForPositiveAnswer();
	}
}

/*------------------------------------------------------------------------------*\
	BccRcpt( mail, sendDataForEachBcc)
		-	announces all Bcc-recipients of given mail to the server
		-	if param sendDataForEachBcc is set, a new mail will be created for each 
			Bcc-recipient, containing only it inside the Bcc-header
\*------------------------------------------------------------------------------*/
void BmSmtp::BccRcpt( BmMail* mail, bool sendDataForEachBcc, 
							 const BmString& headerText) {
	BmAddrList::const_iterator iter;
	const BmAddressList& bccList 
		= mail->IsRedirect() 
			? mail->Header()->GetAddressList( BM_FIELD_RESENT_BCC)
			: mail->Header()->GetAddressList( BM_FIELD_BCC);
	for( iter=bccList.begin(); iter != bccList.end(); ++iter) {
		if (sendDataForEachBcc)
			Mail( mail);
		BmString cmd = BmString("RCPT to:<") << iter->AddrSpec() <<">";
		SendCommand( cmd);
		CheckForPositiveAnswer();
		if (sendDataForEachBcc)
			Data( mail, headerText, iter->AddrSpec());
	}
}

/*------------------------------------------------------------------------------*\
	Data( mail, forBcc)
		-	sends the given mail (the mailtext) to the server
		-	if param forBcc is set, the contained address is set as the mail's
			Bcc-header (only this address)
\*------------------------------------------------------------------------------*/
void BmSmtp::Data( BmMail* mail, const BmString& headerText, BmString forBcc) {
	BmString cmd( "DATA");
	SendCommand( cmd);
	CheckForPositiveAnswer();
	BmString completeHeader;
	if (forBcc.Length()) {
		if (mail->IsRedirect()) {
			// include BCC for current recipient within header so he/she/it can
			// see how this mail got sent to him/her/it:
			completeHeader = BmString("Resent-Bcc: ") << forBcc << "\r\n" 
										<< headerText;
		} else {
			// include BCC for current recipient within header so he/she/it can
			// see how this mail got sent to him/her/it:
			completeHeader = BmString("Bcc: ") << forBcc << "\r\n" << headerText;
		}
	} else
		completeHeader = headerText;
	BmStringIBuf sendBuf( completeHeader);
	sendBuf.AddBuffer( mail->RawText().String()+mail->HeaderLength());
	SendCommand( sendBuf, "", true, true);
	CheckForPositiveAnswer();
}

/*------------------------------------------------------------------------------*\
	Quit( WaitForAnswer)
		-	sends a QUIT to the server, waiting for answer only 
			if WaitForAnswer==true
		-	normally, we wait for an answer, just if we are shutting down
			because of an error we ignore any answer.
		-	the network-connection is always closed
\*------------------------------------------------------------------------------*/
void BmSmtp::Quit( bool WaitForAnswer) {
	BmString cmd("QUIT");
	try {
		SendCommand( cmd);
		if (WaitForAnswer)
			GetAnswer();
	} catch(...) {	}
	Disconnect();
}

/*------------------------------------------------------------------------------*\
	SuggestAuthType()
		-	looks at the auth-types supported by the server and selects the most 
			secure of those that is supported by Beam.
\*------------------------------------------------------------------------------*/
BmString BmSmtp::SuggestAuthType() const {
	if (mSupportedAuthTypes.IFindFirst( BmSmtpAccount::AUTH_LOGIN) != B_ERROR)
		return BmSmtpAccount::AUTH_LOGIN;
	else if (mSupportedAuthTypes.IFindFirst( 
		BmSmtpAccount::AUTH_PLAIN
	) != B_ERROR)
		return BmSmtpAccount::AUTH_PLAIN;
	else
		return "";
}
